;; -*- no-byte-compile: t; lexical-binding: nil -*-
(define-package "ada-ts-mode" "20260502.1153"
  "Major mode for Ada using Tree-sitter."
  '((emacs "29.1"))
  :url "https://github.com/brownts/ada-ts-mode"
  :commit "901b7ce589cfdb3ad4caeb68a97f7280a8383023"
  :revdesc "901b7ce589cf"
  :keywords '("ada" "languages" "tree-sitter")
  :authors '(("Troy Brown" . "brownts@troybrown.dev"))
  :maintainers '(("Troy Brown" . "brownts@troybrown.dev")))
